from typing import List

from . import install

help = "Update packages to its compatible version"
options = [
    {
        "name": ["-g", "--global"],
        "dest": "global_",
        "action": "store_true",
        "help": "Update packages in the global folder"
    },
    {
        "name": ["--latest"],
        "action": "store_true",
        "help": "Install the latest version even if the new version is not compatible"
    },
    {
        "name": "PACKAGE",
        "nargs": "*",
        "help": "Package name"
    }
]

def run(ns):
    from .. import venv
    
    if ns.global_:
        if ns.PACKAGE:
            packages = ns.PACKAGE
        else:
            packages = list(venv.iter_global_packages())
        install.install_global(packages, upgrade=True, latest=ns.latest)
    else:
        vv = venv.get_current_venv()
        with vv.activate():
            update_local(ns.PACKAGE, latest=ns.latest)
        
def update_local(packages: List[str], latest: bool = False):
    """Update local packages.
    
    :arg packages: A list of packages that shoud be updated.
    :arg latest: Whether to upgrade to the latest version.
    """
    from .. import dependency, pip_api

    dev_requires = list(dependency.get_dev_requires())
    prod_requires = list(dependency.get_prod_requires())
    dev_packages = set(r.name for r in dev_requires)
    prod_packages = set(r.name for r in prod_requires)

    if not packages:
        packages = [r.name for r in dev_requires + prod_requires if r.marker is None or r.marker.evaluate()]

    if not packages:
        return

    names = [dependency.spec_to_pkg(i) for i in packages]
    missing = set(names) - dev_packages - prod_packages
    if missing:
        raise Exception(f"Some packages are not installed: {', '.join(missing)}")

    pip_api.install([*packages, *pip_api.freeze(include=dev_packages | prod_packages, exclude=set(names))], upgrade=True, latest=latest)
    infos = pip_api.show(names)

    dev_installed = {}
    prod_installed = {}
    for info, pkg in zip(infos, names):
        if pkg in dev_packages:
            dev_installed[pkg] = info.version
        if pkg in prod_packages:
            prod_installed[pkg] = info.version

    dev_result = dependency.add_dev(dev_installed)
    prod_result = dependency.add_prod(prod_installed)

    if prod_result.incompat_update:
        pip_api.install_editable()

    if any(r.dirty for r in [dev_result, prod_result]):
        dependency.update_lock()



import re
import importlib
import importlib.resources

from typing import Iterator

def old_iter_files() -> Iterator[str]:
    for name in importlib.resources.contents(__name__):
        if re.match(r"[a-z]\w+\.py", name):
            yield name

def iter_files() -> Iterator[str]:
    try:
        dir = importlib.resources.files(__name__)
    except AttributeError:
        yield from old_iter_files()
    else:
        for file in dir.iterdir():
            if file.is_file() and file.name.endswith(".py"):
                yield file.name

#: List of builtin command names.
names = [
    filename.partition(".")[0] for filename in iter_files() if filename != "__init__.py"
]

def get_modules():
    """Get module instances.
    
    :return: A ``command_name -> module`` map.
    :rtype: dict[str, module]
    """
    modules = {}
    for name in names:
        modules[name] = importlib.import_module(".{}".format(name), __name__)
    return modules
    



from __future__ import annotations

import os
import re
import shutil
import sysconfig
import venv
from contextlib import contextmanager
from pathlib import Path
from typing import Iterable, List

from .execute import execute

def get_script_folder(base):
    if os.name == "nt":
        return os.path.join(base, "Scripts")
    return os.path.join(base, "bin")
    
#: Absolute path to the global package venv folder ``~/.vpip/pkg_venvs``
GLOBAL_FOLDER: str = os.path.normpath(os.path.expanduser("~/.vpip/pkg_venvs"))

#: These packages are pre-installed by vpip. They are excluded from the lock file. You can update them via ``update_venv`` command.
PREINSTALLED_PACKAGES: List[str] = ["pip", "wheel"]

class GlobalScriptFolderGetter:
    """Return a list of folders. Which are used to write global scripts.

    Following paths are checked against env variable ``PATH``::

        '~/.local/bin'
        '~/bin'
        sysconfig.get_path('scripts')

    They are only returned if included in ``PATH``.
    """
    def __call__(self) -> Iterable[Path]:
        folders = set([
            Path("~/.local/bin").expanduser(),
            Path("~/bin").expanduser(),
            Path(sysconfig.get_path("scripts"))
        ])
        cache = []
        paths = [Path(p) for p in os.environ["PATH"].split(os.pathsep)]
        for path in paths:
            if path in folders:
                yield path
                cache.append(path)
        if not cache:
            raise Exception('Cannot find a valid scripts folder in env variable PATH')
        
get_global_script_folders = GlobalScriptFolderGetter()

def get_global_folder(pkg_name):
    """Get global venv folder for a package.
    
    :arg str pkg_name: Package name.
    :rtype: str
    """
    return os.path.join(GLOBAL_FOLDER, pkg_name)
    
def iter_global_packages():
    """Iterate through all venv folders for globally installed packages.
    
    :rtype: Iterator[str]
    """
    for dir in Path(GLOBAL_FOLDER).iterdir():
        yield dir.name
    
def get_path_without_venv(path, venv_dir):
    if not venv_dir:
        return path
    return os.pathsep.join(
        p for p in re.split(rf"\s*{os.pathsep}\s*", path)
        if not p.startswith(venv_dir))
                    
def get_current_venv():
    """Get the :class:`Venv` instance pointing to ``./.venv``.
    
    :rtype: Venv
    """
    return Venv(".venv")
    
def get_global_pkg_venv(pkg):
    """Get the :class:`Venv` instance pointing the venv folder of the global
    installed package.
    
    :arg str pkg: Package name. It would also be used as the folder name.
    :rtype: Venv
    """
    return Venv(get_global_folder(pkg))

def get_global_tmp_venv() -> Venv:
    """Get the :class:`Venv` instance with a temporary name."""
    from time import time
    return Venv(get_global_folder(f"tmp-{time()}"))
    
class Builder(venv.EnvBuilder):
    """An environment builder that could be used inside a venv.
    
    It also upgrades pip to the latest version after installed.
    """
    def ensure_directories(self, env_dir):
        context = super().ensure_directories(env_dir)
        current_venv = os.environ.get("VIRTUAL_ENV")
        if current_venv:
            clean_path = get_path_without_venv(os.environ["PATH"], current_venv)
            # find executable that is not in the current virtual env
            # https://github.com/python/cpython/blob/cd449806fac1246cb7b4d392026fe6986ec01fb7/Lib/venv/__init__.py#L113-L116
            executable = shutil.which("python", path=clean_path)
            dirname, exename = os.path.split(executable)
            context.executable = executable
            context.python_dir = dirname
            context.python_exe = exename
        return context
        
    def post_setup(self, context):
        # update pip to latest
        execute([context.env_exe, "-Im", "pip", "install", "-U", *PREINSTALLED_PACKAGES])        
        

class Venv:
    """A helper class that is associated to a venv folder. It allows you to
    easily activate/deactivate the venv.
    """
    def __init__(self, env_dir):
        """
        :arg str env_dir: The target venv folder.
        """
        self.env_dir = os.path.abspath(env_dir)
        self.old_env_dir = os.environ.get("VIRTUAL_ENV")
        
        self.path = "{}{}{}".format(
            get_script_folder(self.env_dir),
            os.pathsep,
            get_path_without_venv(os.environ["PATH"], self.old_env_dir)
        )
        self.old_path = os.environ["PATH"]
        
    def exists(self):
        """Check if the folder exists.
        
        :rtype: bool
        """
        return os.path.exists(self.env_dir)
    
    @contextmanager
    def activate(self, auto_create=False):
        """Activate the venv. Update PATH and VIRTUAL_ENV environment variables.
        
        :arg bool auto_create: If True then automatically create the venv when
            the folder doesn't exist.
            
        This function can be used as a context manager that will
        :meth:`deactivate` when exited.
        """
        try:
            if not self.exists():
                if auto_create:
                    self.create()
                else:
                    raise Exception("venv folder doesn't exists")
            os.environ["PATH"] = self.path
            os.environ["VIRTUAL_ENV"] = self.env_dir
            yield
        finally:
            self.deactivate()
        
    def deactivate(self):
        """Deactivate the venv."""
        os.environ["PATH"] = self.old_path
        if self.old_env_dir:
            os.environ["VIRTUAL_ENV"] = self.old_env_dir
        elif "VIRTUAL_ENV" in os.environ:
            del os.environ["VIRTUAL_ENV"]
    
    def create(self):
        """Create the venv."""
        print("building venv at {}".format(self.env_dir))
        Builder(with_pip=True).create(self.env_dir)
        
    def destroy(self):
        """Destroy the venv. Remove the venv folder."""
        shutil.rmtree(self.env_dir)



def test_pypi_compatible():
    from vpip.pypi import is_compatible
    from packaging.version import parse as p
    assert is_compatible(p("0.1.0"), p("0.2.0")) is False
    assert is_compatible(p("1.1.0"), p("1.2.0")) is True
    assert is_compatible(p("1.1.0"), p("2.2.0")) is False
    
def test_global_script_folder():
    # make sure the script folder is already in the path
    from vpip.venv import get_global_script_folders
    import os
    import pathlib
    paths = set(pathlib.Path(p) for p in os.environ["PATH"].split(os.pathsep))
    assert any(f in paths for f in get_global_script_folders())
    
def test_sub_deps_conflict():
    from yamldirs import create_files
    from vpip.commands.install import install_local_first_time
    from vpip.commands.update import update_local
    from vpip import pip_api, venv
    files = """
    requirements.txt: |
        Sphinx~=4.0
        twine~=3.0
    """
    with create_files(files):
        install_local_first_time()
        with venv.get_current_venv().activate():
            update_local(["twine~=4.0"])
            docutils,  = pip_api.show(["docutils"])
            assert docutils.version.startswith("0.17")
    
def test_marker():
    from yamldirs import create_files
    from vpip.commands.install import install_local_first_time
    from vpip.commands.update import update_local
    from vpip import pip_api, venv
    files = """
    requirements.txt: |
        twine~=3.0; python_version < '3'
    """
    with create_files(files):
        install_local_first_time()
        with venv.get_current_venv().activate():
            assert all(r.name != "twine" for r in pip_api.list_())
            update_local([])
            assert all(r.name != "twine" for r in pip_api.list_())

def test_toml():
    from yamldirs import create_files
    from vpip.commands.install import install_local_first_time, install_local
    from vpip import dependency
    files = """
    requirements.txt: |
        twine~=3.0
    """
    with create_files(files):
        install_local_first_time()
        install_local(["requests"])
        u = dependency.TomlUpdater()
        assert u.get_requirements().startswith("requests~=")

def test_vpip_config():
    from yamldirs import create_files
    from vpip import dependency
    files = """
    setup.cfg: |
        [vpip]
        command_fallback = foo

        [vpip.commands]
        test = bar
    """
    with create_files(files):
        config = dependency.get_vpip_config()
        assert config == {"command_fallback": "foo", "commands": {"test": "bar"}}

def test_vpip_config_toml():
    from yamldirs import create_files
    from vpip import dependency
    files = """
    pyproject.toml: |
        [tool.vpip]
        command_fallback = 'foo'
        commands = {test = 'bar'}
    """
    with create_files(files):
        config = dependency.get_vpip_config()
        assert config == {"command_fallback": "foo", "commands": {"test": "bar"}}



from abc import abstractmethod
import configparser
import re
from collections import OrderedDict
from pathlib import Path
from typing import Iterator, List, Union

from configupdater import ConfigUpdater
from packaging.requirements import Requirement
import tomlkit
from tomlkit.toml_file import TOMLFile

LOCK_FILE = "requirements-lock.txt"

def parse_requirements(text) -> Iterator[Requirement]:
    """Parse requirements text.

    FIXME: switch to an external function from pip if possible.
    https://pip.pypa.io/en/stable/reference/requirements-file-format/#requirements-file-format
    """
    for line in get_continued_lines(text):
        # FIXME: handle options?
        line = re.sub(r"(^|\s+)(#|--\w+|-e).+", "", line)
        if not line:
            continue
        if re.match("https?:", line):
            continue
        if line.startswith("."):
            continue
        yield Requirement(line)

def get_continued_lines(text) -> Iterator[str]:
    last_line = ""
    for line in text.split("\n"):
        last_line += line
        if line.endswith("\\"):
            continue
        if last_line.strip():
            yield last_line.strip()
            last_line = ""
    if last_line.strip():
        yield last_line.strip()

def get_dev_requires():
    return parse_requirements(DevUpdater().get_requirements())
    
def get_prod_requires():
    return parse_requirements(get_prod_updater().get_requirements())

def get_all() -> List[Requirement]:
    m = OrderedDict()
    m.update((r.name, r) for r in get_dev_requires())
    m.update((r.name, r) for r in get_prod_requires())
    return list(m.values())
    
class Updater:
    """Dependency updater interface. Extend this class to create a new updater.
    """
    @abstractmethod
    def get_requirements(self):
        """Get requirements string.
        
        :rtype: str
        """
        raise NotImplementedError
        
    @abstractmethod
    def get_spec(self, name, version):
        """Get version specifier.
        
        :arg str name: Installed package name.
        :arg str version: Installed pacakge version.
        :return: Version specifier e.g. ``"foo==0.1.0"``
        :rtype: str
        """
        raise NotImplementedError
        
    @abstractmethod
    def write_requirements(self, lines):
        """Write new requirements to file.
        
        :arg list[str] line: Lines of requirements.
        """
        raise NotImplementedError

class DevUpdater(Updater):
    """Development dependency (requirements.txt) updater."""
    def __init__(self):
        self.file = Path("requirements.txt")
        
    def get_requirements(self):
        try:
            return self.file.read_text("utf8")
        except OSError:
            pass
        return ""
        
    def get_spec(self, name, version):
        return "{}=={}".format(name, version)
        
    def write_requirements(self, lines):
        with self.file.open("w", encoding="utf8") as f:
            for line in lines:
                f.write(line + "\n")
                
def get_prod_updater():
    if TomlUpdater.available():
        return TomlUpdater()
    if SetupUpdater.available():
        return SetupUpdater()
    return TomlUpdater()

def get_vpip_config() -> dict[str, Union[str, dict]]:
    """Get vpip config dict"""
    try:
        return get_prod_updater().get_vpip_config()
    except OSError:
        return {}

class ProdUpdater(Updater):
    """Production dependency base class"""
    file_path = "" # should be overridden

    def get_spec(self, name, version):
        if not version.startswith("0."):
            version = re.match(r"\d+\.\d+", version).group()
        return "{}~={}".format(name, version)

    @classmethod
    def available(cls):
        return Path(cls.file_path).exists()

    @abstractmethod
    def get_vpip_config(self):
        """Get vpip config dict"""
        raise NotImplementedError
        
class TomlUpdater(ProdUpdater):
    """Production dependency (pyproject.toml) updater."""
    file_path = "pyproject.toml"
    def __init__(self):
        self.file = TOMLFile(self.file_path)
        self.document = None
        
    def read(self):
        if self.document is not None:
            return
        try:
            self.document = self.file.read()
        except OSError:
            return
        
    def get_requirements(self):
        self.read()
        if not self.document:
            return ""
        try:
            items = self.document["project"]["dependencies"]
            return "\n".join(items)
        except KeyError:
            pass
        return ""
        
    def get_name(self):
        self.read()
        return self.document["project"]["name"]
        
    def write_requirements(self, lines):
        if not self.document:
            self.document = tomlkit.document()
        if "project" not in self.document:
            table = tomlkit.table()
            self.document.add("project", table)
        self.document["project"]["dependencies"] = lines
        self.file.write(self.document)

    def get_vpip_config(self):
        self.read()
        if not self.document:
            return {}
        try:
            return self.document["tool"]["vpip"]
        except KeyError:
            pass
        return {}

class SetupUpdater(ProdUpdater):
    """Production dependency (setup.cfg) updater."""
    file_path = "setup.cfg"
    def __init__(self):
        self.file = Path(self.file_path)
        self.file_py = self.file.with_suffix(".py")
        self.config = ConfigUpdater()
        self.indent = "  " # default to 2 spaces indent?
        
    def read(self):
        try:
            text = self.file.read_text("utf8")
        except OSError:
            return
        self.indent = detect_indent(text) or self.indent
        self.config.read_string(text)
        
    def get_requirements(self):
        self.read()
        try:
            return self.config.get("options", "install_requires").value
        except configparser.Error:
            pass
        return ""
        
    def get_name(self):
        self.read()
        return self.config.get("metadata", "name").value
        
    def write_requirements(self, lines):
        if "options" not in self.config:
            self.config.add_section("options")
        if "install_requires" not in self.config["options"]:
            self.config["options"]["install_requires"] = None
        self.config["options"]["install_requires"].set_values(lines)
        self.file.write_text(str(self.config).replace("\r", ""), encoding="utf8")
        if not self.file_py.exists():
            self.file_py.write_text("\n".join([
                "from setuptools import setup",
                "setup()"
            ]), encoding="utf8")

    def get_vpip_config(self):
        self.read()
        result = {}
        try:
            for key in self.config:
                if key == "vpip":
                    for value in self.config[key]:
                        # FIXME: the value is always a string
                        result[value] = self.config[key][value].value
                elif key.startswith("vpip."):
                    n_result = create_nest_dict(result, key.split(".")[1:])
                    for value in self.config[key]:
                        n_result[value] = self.config[key][value].value
            return result
        except KeyError:
            pass
        return {}

def create_nest_dict(d, keys):
    """Create nested dict from keys"""
    for key in keys:
        if key not in d:
            d[key] = {}
        d = d[key]
    return d
        
class UpdateDependencyResult:
    def __init__(self):
        self.dirty = False
        self.add = 0
        self.update = 0
        self.remove = 0
        self.incompat_update = 0
        
def update_dependency(updater, added=None, removed=None):
    """Update dependency and save.
    
    :arg Updater updater: An Updater instance.
    :arg dict added: A ``pkg_name -> version`` map. Added packages.
    :arg list[str] removed: A list of package name. Removed packages.
    """
    added = added or {}
    removed = set(removed or [])
    output = []
    result = UpdateDependencyResult()
    for require in parse_requirements(updater.get_requirements()):
        if require.name in added:
            result.dirty = True
            result.update += 1
            version = added.pop(require.name)
            if version not in require.specifier:
                result.incompat_update += 1
            spec = updater.get_spec(require.name, version)
            if require.marker:
                spec += ";{}".format(require.marker)
            output.append(spec)
        elif require.name in removed:
            result.dirty = True
            result.remove += 1
        else:
            output.append(str(require))
            
    for name, version in added.items():
        result.dirty = True
        result.add += 1
        output.append(updater.get_spec(name, version))
        
    if result.dirty:
        output.sort()
        updater.write_requirements(output)
        
    return result
        
def has_lock() -> bool:
    """Detect if there is a lock file (requirements-lock.txt)"""
    return Path(LOCK_FILE).exists()
        
def update_lock():
    """Run ``pip freeze`` and update the lock file"""
    from . import pip_api
    from .venv import PREINSTALLED_PACKAGES
    lines = pip_api.freeze(exclude=PREINSTALLED_PACKAGES)
    Path(LOCK_FILE).write_text("\n".join(lines), encoding="utf8")

def add_dev(packages):
    return update_dependency(DevUpdater(), added=packages)
    
def add_prod(packages, **kwargs):
    return update_dependency(get_prod_updater(), added=packages)

def delete(packages):
    return (
        update_dependency(DevUpdater(), removed=packages),
        update_dependency(get_prod_updater(), removed=packages)
    )
    
def detect_indent(text):
    for line in text.split("\n"):
        match = re.match(r"(\s+)\S", line)
        if match:
            return match.group(1)
    return None
    
def spec_to_pkg(text: str) -> str:
    return Requirement(text).name



"""``pip`` command API."""

import json
import pathlib
import re
from argparse import Namespace
from typing import List, Optional, Container

from packaging.requirements import Requirement
import case_conversion

from .execute import execute

def install(
    packages: List[str],
    install_scripts: str = None,
    upgrade: bool = False,
    latest: bool = False,
    deps: bool = True
) -> List[str]:
    """Install packages and return a list of collected package names.
    
    :arg packages: A list of package name, which may include the version specifier. It can also be a URL.
    :arg install_scripts: Install scripts to a different folder. It uses
        the ``--install-option="--install-scripts=..."`` pip option.
    :arg upgrade: Upgrade package.
    :arg latest: Whether upgrade to the latest version. Otherwise upgrade
        to the compatible version. This option has no effect if ``package``
        includes specifiers.
    :arg deps: Whether to install dependencies.
    """
    cmd = "install"

    if install_scripts:
        cmd += " --install-option \"--install-scripts={}\"".format(install_scripts)
    if upgrade:
        cmd += " -U --upgrade-strategy eager"
    if not deps:
        cmd += " --no-deps"
        
    need_info = []
    for i, pkg in enumerate(packages):
        if upgrade and not pkg.startswith("http") and not latest and not Requirement(pkg).specifier:
            # compatible update. find current version
            need_info.append((i, pkg))
    result = show([v for k, v in need_info])
    if len(result) != len(need_info):
        installed = set(r.name for r in result)
        needed = set(v for k, v in need_info)
        missing = list(needed - installed)
        raise Exception(f"Upgrade error: some packages are not installed: {', '.join(missing)}")
    for info, (i, pkg) in zip(result, need_info):
        packages[i] = f"{pkg}~={get_compatible_version(info.version)}"

    cmd = f"{cmd} {' '.join(packages)}"
    collected = []
    for line in execute_pip(cmd, capture=True):
        print(line, end="")
        match = re.match("Installing collected packages:(.+)", line, re.I)
        if match:
            collected = [p.strip() for p in match.group(1).split(",")]
    return collected
    
def install_requirements(file="requirements.txt"):
    """Install ``requirements.txt`` file."""
    execute_pip("install -r {}".format(file))
    
def install_editable():
    """Install the current cwd as editable package."""
    setup = pathlib.Path("setup.py")
    if setup.exists():
        execute_pip("install -e .")
    
def uninstall(packages):
    """Uninstall packages.
    
    :arg list[str] package: Package name.
    """
    if not packages:
        return
    execute_pip("uninstall -y {}".format(" ".join(packages)))
    
def show(packages, verbose=False):
    """Get package information.
    
    :arg list[str] packages: A list of package name.
    :arg bool verbose: Whether to return verbose info.
    :return: A list of namespace objects holding the package information.
    :rtype: list[Namespace]
    
    This function uses ``pip show`` under the hood. Property name is generated
    by :func:`case_conversion.snakecase`.
    """
    if not packages:
        return []
        
    cmd = "show"
    if verbose:
        cmd += " --verbose"
        
    result = []
    ns = Namespace()
    last_name = None
    
    for line in execute_pip("{} {}".format(cmd, " ".join(packages)), True):
        if line.startswith("---"):
            result.append(ns)
            ns = Namespace()
            continue
            
        match = re.match(r"([\w-]+):\s*(.*)", line)
        if match:
            name, value = match.groups()
            name = case_conversion.snakecase(name)
            value = value.strip()
            setattr(ns, name, value)
            last_name = name
            continue
            
        match = re.match(r"\s+(\S.*)", line)
        if match and last_name:
            value = getattr(ns, last_name) + "\n" + match.group(1).strip()
            setattr(ns, last_name, value)
            continue
            
    result.append(ns)
    return result
    
def list_(not_required=False, format="json"):
    """List installed packages.
    
    :rtype: list[argparse.Namespace]
    """
    cmd = "list --local --exclude-editable"
    if not_required:
        cmd += " --not-required"
    cmd += " --format {}".format(format)
    lines = []
    for line in execute_pip(cmd, capture=True):
        lines.append(line)
    # Note: only the first line is json, other lines include pip upgrade info, etc
    return [create_ns_from_dict(item) for item in json.loads(lines[0])]

def freeze(include: Optional[Container[str]] = None, exclude: Optional[Container[str]] = None) -> List[str]:
    """List installed packages in ``pip freeze`` format (``my_pkg==1.2.3``).

    :arg include: If defined, only returns specified packages.
    :arg exclude: If defined, exclude specified packages.
    """
    result = []
    for p in list_():
        if include is not None and p.name not in include:
            continue
        if exclude is not None and p.name in exclude:
            continue
        result.append(f"{p.name}=={p.version}")
    return result
    
def create_ns_from_dict(d):
    """Create a namespace object from a dict.
    
    :arg dict d: Dictionary.
    :rtype: argparse.Namespace
    """
    ns = Namespace()
    for key, value in d.items():
        setattr(ns, key, value)
    return ns

def execute_pip(cmd, capture=False):
    """Run pip command.
    
    :arg str cmd: ``pip`` command. It would be prefixed with ``python -m pip``.
    :arg bool capture: Whether to capture output.
    """
    prefix = "python "
    if capture:
        prefix += "-X utf8 "
    prefix += "-m pip "
    if capture:
        prefix += "--no-color "
    return execute(prefix + cmd, capture)
    
def get_compatible_version(version):
    """Return the compatible version.
    
    :arg str version: Version string.
    :return: The compatible version which could be used as ``~={compatible_version}``.
    :rtype: str

    Suppose the version string is ``x.y.z``:
    
    * If ``x`` is zero then return ``x.y.z``.
    * Otherwise return ``x.y``.
    """
    if version.startswith("0."):
        return version
    return ".".join(version.split(".")[:2])
    
